import numpy as np
import pyspm as spm
from scipy.special import erf


def spm_smooth(input_vol, kernel):
    if kernel.size == 1:
        kernel = np.array([kernel, kernel, kernel])

    smoothed_vol = smooth1(input_vol, kernel)

    return smoothed_vol


def smooth1(input_vol, kernel):
    vox = np.array([1, 1, 1])

    kernel = kernel / vox
    kernel1 = kernel / ((8 * np.log(2)) ** .5)

    x = int(np.round(6 * kernel1[0]))
    x = np.array(range(-x, x + 1), ndmin=2)
    x = spm_smoothkern(kernel[0], x, 1)
    x = x / np.sum(x)

    y = int(np.round(6 * kernel1[1]))
    y = np.array(range(-y, y + 1), ndmin=2)
    y = spm_smoothkern(kernel[1], y, 1)
    y = y / np.sum(y)

    z = int(np.round(6 * kernel1[2]))
    z = np.array(range(-z, z + 1), ndmin=2)
    z = spm_smoothkern(kernel[2], z, 1)
    z = z / np.sum(z)

    i = (x.size - 1) / 2
    j = (y.size - 1) / 2
    k = (z.size - 1) / 2

    smoothed_vol = np.zeros(input_vol.shape, order='F')
    input_vol = np.array(input_vol, order='F')
    smoothed_vol = spm.conv_vol(input_vol, smoothed_vol, x, y, z, np.array([-i, -j, -k], ndmin=2))

    return smoothed_vol


def spm_smoothkern(fwhm, x, t=1):
    eps = np.finfo(float).eps
    s = (fwhm / (8 * np.log(2)) ** .5) ** 2 + eps

    if t == 0:

        w1 = 1 / ((2 * s) ** .5)
        krn = 0.5 * (erf(w1 * (x + 0.5)) - erf(w1 * (x - 0.5)))
        krn[krn < 0] = 0

    else:

        w1 = 0.5 * (2 / s) ** .5
        w2 = -0.5 / s
        w3 = (s / 2 / np.pi) ** .5
        krn = 0.5 * (erf(w1 * (x + 1)) * (x + 1) + erf(w1 * (x - 1)) * (x - 1) - 2 * erf(w1 * x) * x) \
              + w3 * (np.exp(w2 * (x + 1) ** 2) + np.exp(w2 * (x - 1) ** 2) - 2 * np.exp(w2 * x ** 2))

        krn[krn < 0] = 0

    return krn



# Rigid body reslicing of images
# FORMAT spm_reslice(P,flags)
#
# P      - matrix or cell array of filenames {one string per row}
#          All operations are performed relative to the first image.
#          ie. Coregistration is to the first image, and resampling
#          of images is into the space of the first image.
#
# flags  - a structure containing various options.  The fields are:
#
#         mask   - mask output images (true/false) [default: true]
#                  To avoid artifactual movement-related variance the
#                  realigned set of images can be internally masked, within
#                  the set (i.e. if any image has a zero value at a voxel
#                  than all images have zero values at that voxel). Zero
#                  values occur when regions 'outside' the image are moved
#                  'inside' the image during realignment.
#
#         mean   - write mean image (true/false) [default: true]
#                  The average of all the realigned scans is written to
#                  an image file with 'mean' prefix.
#
#         interp - the B-spline interpolation method [default: 1]
#                  Non-finite values result in Fourier interpolation. Note
#                  that Fourier interpolation only works for purely rigid
#                  body transformations. Voxel sizes must all be identical
#                  and isotropic.
#
#         which  - values of 0, 1 or 2 are allowed [default: 2]
#                  0   - don't create any resliced images.
#                        Useful if you only want a mean resliced image.
#                  1   - don't reslice the first image.
#                        The first image is not actually moved, so it may
#                        not be necessary to resample it.
#                  2   - reslice all the images.
#                  If which is a 2-element vector, flags.mean will be set
#                  to flags.which(2).
#
#         wrap   - three values of either 0 or 1, representing wrapping in
#                  each of the dimensions. For fMRI, [1 1 0] would be used.
#                  For PET, it would be [0 0 0]. [default: [0 0 0]]
#
#         prefix - prefix for resliced images [default: 'r']
#
# __________________________________________________________________________
#
# The spatially realigned images are written to the original subdirectory
# with the same (prefixed) filename. They are all aligned with the first.
#
# Inputs:
# A series of images conforming to SPM data format (see 'Data Format'). The
# relative displacement of the images is stored in their header.
#
# Outputs:
# The routine uses information in their headers and writes the realigned
# image files to the same subdirectory with a prefix.
# __________________________________________________________________________
# Copyright (C) 1999-2011 Wellcome Trust Centre for Neuroimaging
#
# John Ashburner
# $Id: spm_reslice.m 5929 2014-03-27 14:47:40Z guillaume $
# __________________________________________________________________________
# Adopted for OpenNFT by Yury Koush and John Ashburner.
# Copyright (C) 2016-2019 OpenNFT.org
#
# __________________________________________________________________________
#
# The headers of the images contain a 4x4 affine transformation matrix 'M',
# usually affected by the `realignment' and `coregistration' modules.
# What these matrices contain is a mapping from the voxel coordinates
# (x1,x2,x3) (where the first voxel is at coordinate (1,1,1)), to
# coordinates in millimeters (y1,y2,y3).
#
# y1 = m[0][0] * x1 + m[0][1] * x2 + (m[0][2] * x3 + m[0][3])
# y2 = m[1][0] * x1 + m[1][1] * x2 + (m[1][2] * x3 + m[1][3])
# y3 = m[2][0] * x1 + m[2][1] * x2 + (m[2][2] * x3 + m[2][3])
#
# Assuming that image1 has a transformation matrix M1, and image2 has a
# transformation matrix M2, the mapping from image1 to image2 is: M2\M1
# (ie. from the coordinate system of image1 into millimeters, followed
# by a mapping from millimeters into the space of image2).
#
# Several spatial transformations (realignment, coregistration,
# normalisation) can be combined into a single operation (without the
# necessity of resampling the images several times).
# __________________________________________________________________________
#
# Refs:
#
# Friston KJ, Williams SR, Howard R Frackowiak RSJ and Turner R (1995)
# Movement-related effect in fMRI time-series.  Mag. Res. Med. 35:346-355
#
# W. F. Eddy, M. Fitzgerald and D. C. Noll (1996) Improved Image
# Registration by Using Fourier Interpolation. Mag. Res. Med. 36(6):923-931
#
# R. W. Cox and A. Jesmanowicz (1999)  Real-Time 3D Image Registration
# for Functional MRI. Mag. Res. Med. 42(6):1014-1018
# __________________________________________________________________________

import numpy as np
import pyspm as spm
from rtspm.errors import SpmError


def spm_reslice(r, flags):
    msk = []
    count = []
    integral = []
    v0 = []
    r0_dim = np.array(r[0]['dim'])
    r0_mat = np.array(r[0]['mat'])

    if int(flags['mask']) or int(flags['mean']):

        temp_x1 = np.transpose(np.array(range(1, r0_dim[0] + 1), ndmin=2))
        x1 = np.tile(temp_x1, (1, r0_dim[1]))
        temp_x2 = np.transpose(np.array(range(1, r0_dim[1] + 1), ndmin=2))
        x2 = np.transpose(np.tile(temp_x2, (1, r0_dim[1])))

        if int(flags['mean']):
            count = np.zeros(r0_dim)
            integral = np.zeros(r0_dim)

        if int(flags['mask']):
            msk = [[] for _ in range(r0_dim[2])]  # [None]*P['dim'][0][2]

        for x3 in range(0, r0_dim[2]):
            tmp = np.zeros((r0_dim[0], r0_dim[1]))
            for i in range(0, len(r)):
                ri_dim = r[i]['dim'][:3]
                ri_mat = r[i]['mat']

                try:
                    tmp_division = np.linalg.solve(r0_mat, ri_mat)
                except np.linalg.LinAlgError as err:
                    raise SpmError("R0 and R1 division error in reslice") from err

                temp_tmp, y1, y2, y3 = get_mask(np.linalg.inv(tmp_division), x1, x2, x3 + 1, ri_dim, flags['wrap'])
                tmp += temp_tmp

            if int(flags['mask']):
                msk[x3] = np.argwhere(tmp.reshape(tmp.size, 1) != len(r))[:, 0]

            if int(flags['mean']):
                count[:, :, x3] = tmp

    x1, x2 = np.mgrid[1:r0_dim[0] + 1, 1:r0_dim[1] + 1]

    temp_d = np.array([1, 1, 1]) * int(flags['interp'])
    d = np.hstack((temp_d.T, np.squeeze(flags['wrap'])))
    d = np.array(d, ndmin=2).T
    r0_mat = np.array(r[0]['mat'])

    for i in range(1, len(r)):  # range(0,P.size)

        ri_dim = r[i]['dim'][:3]
        ri_mat = r[i]['mat']

        if (i > 1 and int(flags['which']) == 1) or int(flags['which']) == 2:
            write_vol = 1
        else:
            write_vol = 0
        if write_vol or int(flags['mean']):
            read_vol = 1
        else:
            read_vol = 0

        if read_vol:

            v = np.zeros(r0_dim)
            for x3 in range(0, r0_dim[2]):
                try:
                    tmp_division = np.linalg.solve(r0_mat, ri_mat)
                except np.linalg.LinAlgError as err:
                    raise SpmError("R0 and R1 division error in reslice") from err

                tmp, y1, y2, y3 = get_mask(np.linalg.inv(tmp_division), x1, x2, x3 + 1, ri_dim, flags['wrap'])

                out_vol = spm.bsplins(r[i]['C'], y1, y2, y3, d)

                if int(flags['mean']):
                    integral[:, :, x3] += nan_to_zero(v[:, :, x3])

                if int(flags['mask']):
                    tmp = out_vol
                    tmp[msk[x3]] = 0
                    out_vol = tmp

                v[:, :, x3] = out_vol.reshape(y1.shape)

            if write_vol:
                v0 = v

    return v0


def get_mask(m, x1, x2, x3, dim, wrp):
    tiny = 5e-2  # From spm_vol_utils.cpp
    y1 = m[0][0] * x1 + m[0][1] * x2 + (m[0][2] * x3 + m[0][3])
    y2 = m[1][0] * x1 + m[1][1] * x2 + (m[1][2] * x3 + m[1][3])
    y3 = m[2][0] * x1 + m[2][1] * x2 + (m[2][2] * x3 + m[2][3])
    mask = np.array([True] * y1.size).reshape(y1.shape)
    if wrp[0] == 0:
        mask = np.logical_and(np.logical_and(mask, (y1 >= (1 - tiny))), (y1 <= (dim[0] + tiny)))
    if wrp[1] == 0:
        mask = np.logical_and(np.logical_and(mask, (y2 >= (1 - tiny))), (y2 <= (dim[1] + tiny)))
    if wrp[2] == 0:
        mask = np.logical_and(np.logical_and(mask, (y3 >= (1 - tiny))), (y3 <= (dim[2] + tiny)))

    return mask, y1, y2, y3


def nan_to_zero(vi):
    return np.nan_to_num(vi, copy=True, nan=0, posinf=0, neginf=0)



# Estimation of within modality rigid body movement parameters
# FORMAT P = realign(P,flags)
#
# P     - char array of filenames
#         All operations are performed relative to the first image.
#         ie. Coregistration is to the first image, and resampling
#         of images is into the space of the first image.
#         For multiple sessions, P should be a cell array, where each
#         cell should be a matrix of filenames.
#
# flags - a structure containing various options.  The fields are:
#         quality  - Quality versus speed trade-off.  Highest quality (1)
#                    gives most precise results, whereas lower qualities
#                    gives faster realignment.
#                    The idea is that some voxels contribute little to
#                    the estimation of the realignment parameters.
#                    This parameter is involved in selecting the number
#                    of voxels that are used.
#
#         fwhm     - The FWHM of the Gaussian smoothing kernel (mm) applied
#                    to the images before estimating the realignment
#                    parameters.
#
#         sep      - the default separation (mm) to sample the images.
#
#         rtm      - Register to mean.  If field exists then a two pass
#                    procedure is to be used in order to register the
#                    images to the mean of the images after the first
#                    realignment.
#
#         wrap     - Directions in the volume whose values should wrap
#                    around in. For example, in MRI scans, the images wrap
#                    around in the phase encode direction, so (e.g.) the
#                    subject's nose may poke into the back of the subject's
#                    head.
#
#         PW       -  a filename of a weighting image (reciprocal of
#                    standard deviation).  If field does not exist, then
#                    no weighting is done.
#
#         interp   - B-spline degree used for interpolation
#
#         graphics - display coregistration outputs
#                    default: ~spm('CmdLine')
# __________________________________________________________________________
#
# If no output argument, then an updated voxel to world matrix is written
# to the headers of the images (a .mat file is created for 4D images).
# The details of the transformation are displayed in the results window as
# plots of translation and rotation.
# A set of realignment parameters are saved for each session, named:
# rp_*.txt.
# __________________________________________________________________________
#
# Voxel to world mapping:
#
# These are simply 4x4 affine transformation matrices represented in the
# NIFTI headers (see http://nifti.nimh.nih.gov/nifti-1 ).
# These are normally modified by the `realignment' and `coregistration'
# modules.  What these matrices represent is a mapping from the voxel
# coordinates (x1,x2,x3) (where the first voxel is at coordinate (1,1,1)),
# to coordinates in millimeters (y1,y2,y3).
#
# y1 = m[0, 0] * x1 + m[0, 1] * x2 + m[0, 2] * x3 + m[0, 3]
# y2 = m[1, 0] * x1 + m[1, 1] * x2 + m[1, 2] * x3 + m[1, 3]
# y3 = m[2, 0] * x1 + m[2, 1] * x2 + m[2, 2] * x3 + m[2, 3]
#
# Assuming that image1 has a transformation matrix M1, and image2 has a
# transformation matrix M2, the mapping from image1 to image2 is: M2\M1
# (ie. from the coordinate system of image1 into millimeters, followed
# by a mapping from millimeters into the space of image2).
#
# These matrices allow several realignment or coregistration steps to be
# combined into a single operation (without the necessity of resampling the
# images several times).
# __________________________________________________________________________
#
# Reference:
#
# Friston KJ, Ashburner J, Frith CD, Poline J-B, Heather JD & Frackowiak
# RSJ (1995) Spatial registration and normalization of images Hum. Brain
# Map. 2:165-189
# __________________________________________________________________________
# Copyright (C) 1994-2013 Wellcome Trust Centre for Neuroimaging
#
# John Ashburner
# $Id: spm_realign.m 6070 2014-06-26 20:53:39Z guillaume $
#
# SVNid = '$Rev: 6070 $';
#
# P  - a vector of volumes (see spm_vol)
# --------------------------------------------------------------------------
# P(i).mat is modified to reflect the modified position of the image i.
# The scaling (and offset) parameters are also set to contain the
# optimum scaling required to match the images.
# __________________________________________________________________________
#
# Adopted for OpenNFT by Yury Koush and John Ashburner.
# Copyright (C) 2016-2019 OpenNFT.org
#
#
# Real-time computational modifications.
# Note, to speed up the computations,
# 'Coef' are transferred to the spm_reslice_rt, whcih implies
# 1) that exactly the same step in reslice is disabled, and
# 2) that interpolation specified in realign will be 'used'
# by default for the same step of Coef estimation during reslice
# (see smooth_vol() occurences).
# We recommend using the same interpolation for real-time adaptations of
# the realign and reslice functions.
#
# If flag is set to false, the SPM standard realignment and reslicing could
# be reproduced with the high precision. This results in longer ~3-7 sec
# processing of the first volume and a significant reduction of volume
# preprocessing time on each iteration. Relatively long processing of the
# first volume may require a longer first baseline block or alternative
# solutions when higher preprocessing speed is required, e.g. TR = 500ms.

import numpy as np
import pyspm as spm
from rtspm.spm_matrix import spm_matrix
from loguru import logger

NFB_TH_ACC = 0.01
NFB_NR_ITER = 10
SPM_TH_ACC = 1e-8
SPM_NR_ITER = 64
MASK_THRESHOLD = 32


def spm_realign(r, flags, ind_vol, ind_first_vol, a0, x1, x2, x3, deg, b):
    f_nfb = True

    lkp = flags["lkp"]
    # Note: Matlab lkp starts from 1, check input flags
    if lkp[0] == 1:
        lkp -= 1

    if ind_vol == ind_first_vol:

        template_mat = r[0]["mat"][:3, :3]
        skip = (1 / np.sqrt(np.sum(template_mat ** 2, axis=0))) * flags["sep"]
        d = r[0]["dim"][:3]

        # Note: Python random number generation process differs from matlab
        # This is why the result is not identical to Matlab SPM
        rng = np.random.default_rng(0)

        def add_noise(x, k=0.5):
            return x + rng.random(x.shape) * k

        if d[2] < 3:
            lkp = np.array([0, 1, 5])
            x1, x2, x3 = np.mgrid[1:d[0] - 0.5:skip[0], 1:d[1] - 0.5:skip[1], 1:d[2]:skip[2]]
            x1 = add_noise(x1, 0.5)
            x2 = add_noise(x2, 0.5)
        else:
            x1, x2, x3 = np.mgrid[1:d[0] - 0.5:skip[0], 1:d[1] - 0.5:skip[1], 1:d[2] - 0.5:skip[2]]
            x1 = add_noise(x1, 0.5)
            x2 = add_noise(x2, 0.5)
            x3 = add_noise(x3, 0.5)

        x1 = x1.ravel()
        x2 = x2.ravel()
        x3 = x3.ravel()

        # Compute rate of change of chi2 w.r.t changes in parameters(matrix A)
        # ----------------------------------------------------------------------
        v, r[0]["C"] = smooth_vol(r[0], flags["interp"], flags["wrap"], flags["fwhm"])
        temp_d = np.array([1, 1, 1]) * int(flags['interp'])
        deg = np.hstack((temp_d.T, np.squeeze(flags['wrap'])))
        deg = np.array(deg, ndmin=2).T

        g, d_g1, d_g2, d_g3 = spm.bsplins_multi(v, x1, x2, x3, deg)
        a0 = make_a(r[0]["mat"], x1, x2, x3, d_g1, d_g2, d_g3, lkp)
        b = g

    # -Loop over images
    # --------------------------------------------------------------------------
    # control over accuracy and number of iterations
    if f_nfb:
        th_acc = NFB_TH_ACC
        nr_iter = NFB_NR_ITER
    else:
        # SPM defualt
        th_acc = SPM_TH_ACC
        nr_iter = SPM_NR_ITER

    v, r[1]["C"] = smooth_vol(r[1], flags["interp"], flags["wrap"], flags["fwhm"])

    d = np.array(v.shape)
    ss = np.inf
    countdown = -1
    fix_a0 = []
    iteration = 1
    r0_mat = r[0]["mat"]
    r1_mat = r[1]["mat"]

    for iteration in range(1, nr_iter + 1):

        y1, y2, y3 = coords(np.zeros((6, 1)), r0_mat, r1_mat, x1, x2, x3)
        msk = np.nonzero((y1 >= 1) & (y1 <= d[0]) & (y2 >= 1) & (y2 <= d[1]) & (y3 >= 1) & (y3 <= d[2]))
        msk = msk[0]

        if msk.size < MASK_THRESHOLD:
            logger.error('There is not enough overlap in '
                         'the images to obtain a solution. Offending image: "{}"', r[1].name)

        f = spm.bsplins(v, y1[msk], y2[msk], y3[msk], deg)

        a = a0[msk, :]
        b1 = b[msk]
        sc = np.sum(b1) / np.sum(f)
        b1 = b1 - np.squeeze(f * sc)

        if f_nfb:
            if iteration == 1:
                fix_a0 = a0.T @ a0
            soln = np.linalg.solve(fix_a0, (a.T @ b1))
        else:
            soln = np.linalg.solve((a.T @ a), (a.T @ b1))

        p = np.array([0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0], dtype=float)
        p[lkp] += soln.T
        r1_mat = np.linalg.solve(spm_matrix(p), r1_mat)

        pss = ss
        ss = np.sum(b1 ** 2) / b1.size
        if ((pss - ss) / pss < th_acc) and (countdown == -1):
            countdown = 2
        if countdown != -1:
            if countdown == 0:
                break
            countdown -= 1

    r[1]["mat"] = r1_mat
    nr_iter = iteration

    return r, a0, x1, x2, x3, deg, b, nr_iter


def coords(p, m1, m2, x1, x2, x3):
    m = np.linalg.inv(m2) @ np.linalg.inv(spm_matrix(p)) @ m1
    y1 = m[0, 0] * x1 + m[0, 1] * x2 + m[0, 2] * x3 + m[0, 3]
    y2 = m[1, 0] * x1 + m[1, 1] * x2 + m[1, 2] * x3 + m[1, 3]
    y3 = m[2, 0] * x1 + m[2, 1] * x2 + m[2, 2] * x3 + m[2, 3]

    return y1, y2, y3


def smooth_vol(r, hld, wrp, fwhm):
    s = np.sqrt(np.sum(r["mat"][:3, :3] ** 2, axis=0)) ** (-1) * (fwhm / np.sqrt(8 * np.log(2)))

    x = round(6 * s[0])
    x = np.array(range(-x, x + 1), ndmin=2)

    y = round(6 * s[1])
    y = np.array(range(-y, y + 1), ndmin=2)

    z = round(6 * s[2])
    z = np.array(range(-z, z + 1), ndmin=2)

    x = np.exp(-x ** 2 / (2 * s[0] ** 2))
    y = np.exp(-y ** 2 / (2 * s[1] ** 2))
    z = np.exp(-z ** 2 / (2 * s[2] ** 2))

    x = x / np.sum(x)
    y = y / np.sum(y)
    z = z / np.sum(z)

    i = (x.size - 1) / 2
    j = (y.size - 1) / 2
    k = (z.size - 1) / 2

    temp_d = np.array([1, 1, 1], ndmin=2) * int(hld)
    d = np.hstack((temp_d.T, np.array(wrp, ndmin=2)))

    coef = spm.bsplinc(r["Vol"], d)
    coef = coef.reshape(r["dim"])

    v = np.zeros(coef.shape, order='F')
    v = spm.conv_vol(coef, v, x, y, z, np.array([-i, -j, -k], ndmin=2))

    return v, coef


def make_a(m, x1, x2, x3, dg1, dg2, dg3, lkp):
    # Matrix of rate of change of weighted difference w.r.t.parameter changes
    p0 = np.array([0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0], dtype=float)
    a = np.zeros((x1.size, lkp.size))
    for i in range(0, lkp.size):
        pt = p0.copy()
        pt[lkp[i]] = pt[i] + 1e-6
        y1, y2, y3 = coords(pt, m, m, x1, x2, x3)
        tmp = np.sum(np.array([y1 - x1, y2 - x2, y3 - x3]).squeeze() * np.array([dg1, dg2, dg3]), 0,
                     keepdims=True).T / (-1e-6)
        a[:, i] = tmp.T

    return a


def error_message(r):
    print('There is not enough overlap in the images to obtain a solution. Offending image: {:} \n'.format(r.fname))



import numpy as np
from scipy.special import betainc


def spm_inv_b_cdf(f, v, w, tol=1e-8):
    max_it = 1e4

    ad = np.array([2, 2, 2])
    rd = np.max(ad)
    az = np.array([[1, 1],
                   [1, 1],
                   [1, 1]])
    rs = np.max(az)
    xa = np.prod(az, 1) > 1

    x = np.zeros(rs)

    md = 0 <= f <= 1 and v > 0 and w > 0

    x[md and f == 1] = 1

    q = np.nonzero(md and 0 < f < 1)[0]
    if q.size == 0:
        return x

    if xa[0]:
        fq = f
        fq = fq.flatten()
    else:
        fq = f * np.ones((np.max(q.shape), 1))

    if xa[1]:
        vq = v
        vq = vq.flatten()
    else:
        vq = v * np.ones((np.max(q.shape), 1))

    if xa[2]:
        wq = w
        wq = wq.flatten()
    else:
        wq = w * np.ones((np.max(q.shape), 1))

    a = np.zeros((np.max(q.shape), 1))
    fa = a - fq
    b = np.ones((np.max(q.shape), 1))
    fb = b - fq
    i = 0
    xq = a + 0.5
    qq = np.array(range(0, (np.max(q.shape))))

    while qq.size > 0 and i < max_it:
        fxqq = betainc(vq[qq], wq[qq], xq[qq]) - fq[qq]
        mqq = fa[qq] * fxqq > 0

        a[qq[mqq[0]]] = xq[qq[mqq[0]]]
        fa[qq[mqq[0]]] = fxqq[mqq[0]]
        b[qq[not mqq]] = xq[qq[not mqq]]
        fb[qq[not mqq]] = fxqq[not mqq]
        xq[qq] = a[qq] + (b[qq] - a[qq]) / 2
        qq = qq[((b[qq] - a[qq]) > tol)[0]]

        i += 1

    x[q] = xq

    return x


"""
Building Python extension modules via pybind11

The script can be used from Poetry pyproject.toml:

[tool.poetry]
build = "buildext.py"
...

poetry install

Or directly via distutils 'build' command:

python buildext.py build
python buildext.py build --debug  # for debug symbols

"""

import sys
from pathlib import Path
import shutil

from pybind11.setup_helpers import Pybind11Extension, build_ext

from setuptools import setup
from distutils.core import Distribution


MODULE_NAME = 'pyspm'

ROOT_PATH = Path(__name__).parent
SOURCE_DIR = ROOT_PATH / 'src'
SOURCES = sorted(map(str, SOURCE_DIR.glob('*.cpp')))

DEFINE_MACROS = []

if sys.platform == 'win32':
    DEFINE_MACROS.append(
        ('SPM_WIN32', None),
    )

ext_modules = [
    Pybind11Extension(MODULE_NAME, SOURCES, define_macros=DEFINE_MACROS),
]


def build(setup_kwargs: dict):
    setup_kwargs.update({
        "ext_modules": ext_modules,
        "cmdclass": {"build_ext": build_ext}
    })


def copy_lib_files(dist: Distribution):
    build_cmd = dist.get_command_obj('build')
    lib_dir = ROOT_PATH / build_cmd.build_lib

    for fpath in lib_dir.iterdir():
        shutil.copy2(fpath, ROOT_PATH)


def main():
    """Minimal setup to build pybind11 extension modules with debug symbols

    Usage::

        python buildext.py build --debug

    """

    setup_kwargs = {
        'name': MODULE_NAME,
    }

    build(setup_kwargs)
    dist = setup(**setup_kwargs)
    copy_lib_files(dist)


if __name__ == '__main__':
    main()



import numpy as np
from rtspm.spm_matrix import spm_matrix


def spm_imatrix(m):
    r = m[0:3, 0:3]
    c = np.linalg.cholesky(r.T @ r)

    p = np.append(m[0:3, 3].T, np.zeros(3, ))
    p = np.append(p, np.diag(c).T)
    p = np.append(p, np.zeros(3, ))

    if np.linalg.det(r) < 0:
        p[6] = -p[6]

    c = np.linalg.solve(np.diag(np.diag(c)), c)
    p[9:12] = c.flatten()[[3, 6, 7]]
    r0 = spm_matrix(np.append(np.zeros(6, ), p[6:12]))
    r0 = r0[0:3, 0:3]
    r1 = np.linalg.solve(r0.T, r.T).T

    def rang(x):
        return np.minimum(np.maximum(x, -1), 1)

    p[4] = np.arcsin(rang(r1[0, 2]))
    if (np.abs(p[4]) - np.pi / 2) ** 2 < 1e-9:
        p[3] = 0
        p[5] = np.arctan2(-rang(r1[1, 0]), rang(-r1[2, 0] / r1[0, 2]))
    else:
        c = np.cos(p[4])
        p[3] = np.arctan2(rang(r1[1, 2] / c), rang(r1[2, 2] / c))
        p[5] = np.arctan2(rang(r1[0, 1] / c), rang(r1[0, 0] / c))

    return p


# -*- coding: utf-8 -*-

from importlib_metadata import version, PackageNotFoundError

try:
    __version__ = version('rtspm')
except PackageNotFoundError:  # pragma: no cover
    __version__ = '0.0.0.dev'


import numpy as np
from rtspm.spm_inv_b_cdf import spm_inv_b_cdf


def spm_inv_t_cdf(f, v):
    ad = np.array([2, 2])
    rd = np.max(ad)
    az = np.array([[1, 1],
                   [1, 1]])
    rs = np.max(az)
    xa = np.prod(az, 1) > 1

    x = np.zeros(rs)

    md = 0 <= f <= 1 and v > 0

    x[md and f == 0] = -np.inf
    x[md and f == 1] = +np.inf

    ml = (md and v == 1)
    if xa[0]:
        mlf = ml
    else:
        mlf = 1

    x[ml] = np.tan(np.pi * (f - 0.5))

    q = np.nonzero(md and f != 0.5 and 1 > f > 0 != v)[0]
    if q.size == 0:
        return x

    xqpos = f > 0.5
    bq = spm_inv_b_cdf(2 * (xqpos - (xqpos * 2 - 1) * f), v / 2, 0.5)
    x[q] = (xqpos * 2 - 1) * np.sqrt(v / bq - v)

    return x


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['rtspm']

package_data = \
{'': ['*']}

install_requires = \
['importlib-metadata>=4.8.1,<5.0.0',
 'loguru>=0.5.3,<0.6.0',
 'numpy>=1.19.2,<2.0.0',
 'scipy>=1.5.4']

setup_kwargs = {
    'name': 'python-rtspm',
    'version': '0.1.4',
    'description': 'Python adaptation of SPM functions for real-time fMRI analysis',
    'long_description': '# python-rtspm\nSPM functions in Python for real-time fMRI\n',
    'author': 'OpenNFT Team',
    'author_email': 'opennft@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.1,<4',
}
from buildext import *
build(setup_kwargs)

setup(**setup_kwargs)
